from promise import Promise
from cdp_client import cdp
from cdp_client.tests import fake_data as data
from copy import copy
import unittest
import mock


class NodeTester(unittest.TestCase):
    def __init__(self, method_name):
        unittest.TestCase.__init__(self, method_name)
        self._connection = None

    def setUp(self):
        self._connection = cdp.Connection("foo", "bar", False)
        self._root_node = copy(data.system_node)
        self._root_node.node.extend([data.app1_node, data.app2_node])

    def tearDown(self):
        del self._connection
        del self._root_node

    def test_node_info(self):
        node = cdp.Node(None, self._connection, data.app1_node)
        self.assertEqual(node._id(), data.app1_node.info.node_id)
        self.assertEqual(node.name(), data.app1_node.info.name)
        self.assertEqual(node.type(), cdp.NodeType.APPLICATION)
        self.assertEqual(node.is_leaf(), True)
        self.assertEqual(node.is_read_only(), True)
        self.assertEqual(node.path(), node.name())

        sub_node = cdp.Node(node, self._connection, data.app2_node)
        self.assertEqual(sub_node.path(), node.path() + '.' + sub_node.name())

    def test_value(self):
        node = cdp.Node(None, self._connection, data.value1_node)
        node._update_value(data.value1)
        self.assertEquals(node.last_value(), data.value1.d_value)

    @mock.patch.object(cdp.Connection, 'send_value')
    def test_value_setter(self, mock_send_value):
        node = cdp.Node(None, self._connection, data.value1_node)
        node.set_value(data.value1.d_value, data.value1.timestamp)
        mock_send_value.assert_called_with(data.value1)

    @mock.patch.object(cdp.Connection, 'send_structure_request')
    def test_child_getter_when_child_is_leaf(self, mock_send_structure_request):
        children = []
        node = cdp.Node(None, self._connection, self._root_node)
        node.child(data.app1_node.info.name).then(lambda n: children.append(n))
        mock_send_structure_request.assert_not_called()
        self.assertEqual(children[0]._id(), data.app1_node.info.node_id)
        self.assertEqual(children[0].name(), data.app1_node.info.name)
        self.assertEqual(children[0].type(), cdp.NodeType.APPLICATION)

    @mock.patch.object(cdp.Connection, 'send_structure_request')
    def test_child_getter_when_child_is_not_leaf(self, mock_send_structure_request):
        children = []
        node = cdp.Node(None, self._connection, self._root_node)
        mock_send_structure_request.return_value = Promise(lambda resolve, reject: resolve(data.app2_node))
        node.child(data.app2_node.info.name).then(lambda n: children.append(n))
        mock_send_structure_request.assert_called_once_with(data.app2_node.info.node_id, node.path() + '.' + data.app2_node.info.name)
        self.assertEqual(children[0]._id(), data.app2_node.info.node_id)
        self.assertEqual(children[0].name(), data.app2_node.info.name)
        self.assertEqual(children[0].type(), cdp.NodeType.APPLICATION)

    @mock.patch.object(cdp.Connection, 'send_structure_request')
    def test_invalid_child_getter(self, mock_send_structure_request):
        errors = []
        node = cdp.Node(None, self._connection, self._root_node)
        node.child("invalid").catch(lambda e: errors.append(e))
        self.assertEqual(len(errors), 1)
        mock_send_structure_request.assert_not_called()

    @mock.patch.object(cdp.Connection, 'send_structure_request')
    def test_children_getter(self, mock_send_structure_request):
        children = []
        node = cdp.Node(None, self._connection, self._root_node)
        mock_send_structure_request.return_value = Promise(lambda resolve, reject: resolve(data.app2_node))
        node.children().then(lambda n: children.extend(n))
        mock_send_structure_request.assert_called_once_with(data.app2_node.info.node_id, node.path() + '.' + data.app2_node.info.name)
        self.assertEquals(len(children), len(self._root_node.node))

    @mock.patch.object(cdp.Connection, 'send_structure_request')
    def test_children_iterator(self, mock_send_structure_request):
        children = []
        node = cdp.Node(None, self._connection, self._root_node)
        mock_send_structure_request.return_value = Promise(lambda resolve, reject: resolve(data.app2_node))
        node.for_each_child(lambda n: children.append(n))
        mock_send_structure_request.assert_called_once_with(data.app2_node.info.node_id, node.path() + '.' + data.app2_node.info.name)
        self.assertEquals(len(children), len(self._root_node.node))

    @mock.patch.object(cdp.Connection, 'send_structure_request')
    def test_structure_subscription(self, mock_send_structure_request):
        def on_change(added, removed):
            nodes_added.extend(added)
            nodes_removed.extend(removed)

        old_children = []
        new_children = []
        nodes_added = []
        nodes_removed = []
        node = cdp.Node(None, self._connection, self._root_node)
        node.subscribe_to_structure_changes(on_change)
        mock_send_structure_request.return_value = Promise(lambda resolve, reject: resolve(data.app2_node))
        node.children().then(lambda n: old_children.extend(n))

        #change the structure - remove and add a node
        node_to_remove = self._root_node.node[0]
        node_to_add = data.app3_node
        self._root_node.node.remove(node_to_remove)
        self._root_node.node.extend([node_to_add])
        mock_send_structure_request.return_value = Promise(lambda resolve, reject: resolve(self._root_node))
        node._update()

        #verify
        mock_send_structure_request.return_value = Promise(lambda resolve, reject: resolve(node_to_add))
        node.children().then(lambda n: new_children.extend(n))
        self.assertEquals(len(old_children), len(new_children))
        found = False
        for child in new_children:
            self.assertFalse(child.name() == nodes_removed[0])
            if child.name() == nodes_added[0]:
                found = True
        self.assertTrue(found)

    def test_structure_unsubscription(self):
        def on_change(added, removed):
            nodes_added.extend(added)
            nodes_removed.extend(removed)

        nodes_added = []
        nodes_removed = []
        node = cdp.Node(None, self._connection, self._root_node)
        node.subscribe_to_structure_changes(on_change)
        node.unsubscribe_from_structure_changes(on_change)

        # change the structure - remove a node
        node_to_remove = self._root_node.node[0]
        self._root_node.node.remove(node_to_remove)
        node._update_structure(self._root_node)

        # verify
        self.assertEquals(nodes_added, [])
        self.assertEquals(nodes_removed, [])

    @mock.patch.object(cdp.Connection, 'send_value_request')
    def test_value_subscription(self, mock_send_value_request):
        def on_change(value, timestamp):
            actual_values.append(value)
            timestamps.append(timestamp)

        actual_values = []
        timestamps = []
        node = cdp.Node(None, self._connection, data.value1_node)
        node.subscribe_to_value_changes(on_change)
        mock_send_value_request.assert_called_once_with(node._id())

        expected_values = [data.value1, data.value2]
        for value in expected_values:
            node._update_value(value)

        self.assertEquals(len(actual_values), len(expected_values))
        self.assertEquals(len(timestamps), len(expected_values))
        for index, value in enumerate(expected_values):
            self.assertEquals(actual_values[index], value.d_value)
            self.assertEquals(timestamps[index], value.timestamp)

    @mock.patch.object(cdp.Connection, 'send_value_request')
    @mock.patch.object(cdp.Connection, 'send_value_unrequest')
    def test_value_unsubscription(self, mock_send_value_unrequest, mock_send_value_request):
        def on_change(value):
            values.append(value)

        values = []
        node = cdp.Node(None, self._connection, data.value1_node)
        node.subscribe_to_value_changes(on_change)
        mock_send_value_request.assert_called_once_with(node._id())
        node.unsubscribe_from_value_changes(on_change)
        mock_send_value_unrequest.assert_called_once_with(node._id())
        node._update_value(data.value1)
        self.assertEquals(len(values), 0)



from cdp_client import cdp
from cdp_client.tests import fake_data
from promise import Promise
from copy import copy
import unittest
import mock


class ClientTester(unittest.TestCase):
    def __init__(self, method_name):
        unittest.TestCase.__init__(self, method_name)
        self._client = None

    def setUp(self):
        self._client = cdp.Client("foo")

    def tearDown(self):
        del self._client

    @mock.patch.object(cdp.NodeTree, 'root_node')
    def test_root_getter(self, mock_root_node):
        nodes = []
        mock_root_node.return_value = Promise(lambda resolve, reject: resolve(cdp.Node(None, self._client._connection, fake_data.system_node)))
        self._client.root_node().then(lambda node: nodes.append(node))
        mock_root_node.assert_called_once_with()
        self.assertEquals(nodes[0]._id(), fake_data.system_node.info.node_id)

    @mock.patch.object(cdp.NodeTree, 'root_node')
    @mock.patch.object(cdp.Node, 'child')
    def test_find_node(self, mock_child, mock_root_node):
        nodes = []
        comp_node = copy(fake_data.comp1_node)
        app_node = copy(fake_data.app1_node)
        app_node.node.extend([comp_node])
        comp_node.node.extend([fake_data.value1_node])
        mock_root_node.return_value = Promise(lambda resolve, reject: resolve(cdp.Node(None, self._client._connection, app_node)))
        mock_child.side_effect = [Promise(lambda resolve, reject: resolve(cdp.Node(None, self._client._connection, comp_node))),
                                  Promise(lambda resolve, reject: resolve(cdp.Node(None, self._client._connection, fake_data.value1_node)))]
        path = '.'.join([app_node.info.name, comp_node.info.name, fake_data.value1_node.info.name])
        self._client.find_node(path).then(lambda node: nodes.append(node))
        self.assertEquals(nodes[0]._id(), fake_data.value1_node.info.node_id)

    @mock.patch.object(cdp.Connection, 'run_event_loop')
    def test_run_event_loop(self, mock_run_event_loop):
        self._client.run_event_loop()
        mock_run_event_loop.assert_called_once_with()

    @mock.patch.object(cdp.Connection, 'close')
    def test_disconnect(self, mock_close):
        self._client.disconnect()
        mock_close.assert_called_once_with()



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: cdp.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='cdp.proto',
  package='',
  syntax='proto2',
  serialized_options=None,
  serialized_pb=_b('\n\tcdp.proto\"\xe9\x01\n\x05Hello\x12\x13\n\x0bsystem_name\x18\x01 \x02(\t\x12\x19\n\x0e\x63ompat_version\x18\x02 \x02(\r:\x01\x31\x12\x1e\n\x13incremental_version\x18\x03 \x02(\r:\x01\x30\x12\x12\n\npublic_key\x18\x04 \x03(\x0c\x12\x11\n\tchallenge\x18\x05 \x01(\x0c\x12\x18\n\x10\x61pplication_name\x18\x06 \x01(\t\x12\x19\n\x11\x63\x64p_version_major\x18\x07 \x01(\r\x12\x19\n\x11\x63\x64p_version_minor\x18\x08 \x01(\r\x12\x19\n\x11\x63\x64p_version_patch\x18\t \x01(\r\"\x8f\x01\n\x0b\x41uthRequest\x12\x0f\n\x07user_id\x18\x01 \x01(\t\x12:\n\x12\x63hallenge_response\x18\x02 \x03(\x0b\x32\x1e.AuthRequest.ChallengeResponse\x1a\x33\n\x11\x43hallengeResponse\x12\x0c\n\x04type\x18\x01 \x01(\t\x12\x10\n\x08response\x18\x02 \x01(\x0c\"\xb0\x01\n#AdditionalChallengeResponseRequired\x12\x0c\n\x04type\x18\x01 \x01(\t\x12\x0e\n\x06prompt\x18\x02 \x01(\t\x12\x41\n\tparameter\x18\x03 \x03(\x0b\x32..AdditionalChallengeResponseRequired.Parameter\x1a(\n\tParameter\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"\xd9\x02\n\x0c\x41uthResponse\x12\x31\n\x0bresult_code\x18\x01 \x01(\x0e\x32\x1c.AuthResponse.AuthResultCode\x12\x13\n\x0bresult_text\x18\x02 \x01(\t\x12T\n&additional_challenge_response_required\x18\x03 \x03(\x0b\x32$.AdditionalChallengeResponseRequired\"\xaa\x01\n\x0e\x41uthResultCode\x12\x0c\n\x08\x65Unknown\x10\x00\x12\x0c\n\x08\x65Granted\x10\x01\x12\"\n\x1e\x65GrantedPasswordWillExpireSoon\x10\x02\x12\x18\n\x14\x65NewPasswordRequired\x10\n\x12\x1d\n\x19\x65InvalidChallengeResponse\x10\x0b\x12\x1f\n\x1b\x65\x41\x64\x64itionalResponseRequired\x10\x0c\"\xa2\x05\n\tContainer\x12%\n\x0cmessage_type\x18\x01 \x01(\x0e\x32\x0f.Container.Type\x12\x15\n\x05\x65rror\x18\x02 \x01(\x0b\x32\x06.Error\x12\x19\n\x11structure_request\x18\x03 \x03(\r\x12!\n\x12structure_response\x18\x04 \x03(\x0b\x32\x05.Node\x12%\n\x0egetter_request\x18\x05 \x03(\x0b\x32\r.ValueRequest\x12&\n\x0fgetter_response\x18\x06 \x03(\x0b\x32\r.VariantValue\x12%\n\x0esetter_request\x18\x07 \x03(\x0b\x32\r.VariantValue\x12!\n\x19structure_change_response\x18\x08 \x03(\r\x12\x1d\n\x15\x63urrent_time_response\x18\t \x01(\x04\x12$\n\x11\x63hild_add_request\x18\n \x03(\x0b\x32\t.ChildAdd\x12*\n\x14\x63hild_remove_request\x18\x0b \x03(\x0b\x32\x0c.ChildRemove\"\x84\x02\n\x04Type\x12\x10\n\x0c\x65RemoteError\x10\x00\x12\x15\n\x11\x65StructureRequest\x10\x01\x12\x16\n\x12\x65StructureResponse\x10\x02\x12\x12\n\x0e\x65GetterRequest\x10\x03\x12\x13\n\x0f\x65GetterResponse\x10\x04\x12\x12\n\x0e\x65SetterRequest\x10\x05\x12\x1c\n\x18\x65StructureChangeResponse\x10\x06\x12\x17\n\x13\x65\x43urrentTimeRequest\x10\x07\x12\x18\n\x14\x65\x43urrentTimeResponse\x10\x08\x12\x14\n\x10\x65\x43hildAddRequest\x10\t\x12\x17\n\x13\x65\x43hildRemoveRequest\x10\n*\x08\x08\x64\x10\x80\x80\x80\x80\x02\"Q\n\x05\x45rror\x12\x0c\n\x04\x63ode\x18\x01 \x02(\r\x12\x0c\n\x04text\x18\x02 \x01(\t\x12\x0f\n\x07node_id\x18\x03 \x01(\r\x12\x11\n\tparameter\x18\x04 \x01(\t*\x08\x08\x64\x10\x80\x80\x80\x80\x02\"\x96\x03\n\x04Info\x12\x0f\n\x07node_id\x18\x01 \x02(\r\x12\x0c\n\x04name\x18\x02 \x02(\t\x12\x1f\n\tnode_type\x18\x03 \x02(\x0e\x32\x0c.CDPNodeType\x12!\n\nvalue_type\x18\x04 \x01(\x0e\x32\r.CDPValueType\x12\x11\n\ttype_name\x18\x05 \x01(\t\x12\x13\n\x0bserver_addr\x18\x06 \x01(\t\x12\x13\n\x0bserver_port\x18\x07 \x01(\r\x12\x10\n\x08is_local\x18\x08 \x01(\x08\x12\r\n\x05\x66lags\x18\t \x01(\r\"\xc2\x01\n\x05\x46lags\x12\t\n\x05\x65None\x10\x00\x12\x0f\n\x0b\x65NodeIsLeaf\x10\x01\x12\x16\n\x12\x65ValueIsPersistent\x10\x02\x12\x14\n\x10\x65ValueIsReadOnly\x10\x04\x12\x14\n\x10\x65NodeIsRemovable\x10\x08\x12\x17\n\x13\x65NodeCanAddChildren\x10\x10\x12\x14\n\x10\x65NodeIsRenamable\x10 \x12\x13\n\x0f\x65NodeIsInternal\x10@\x12\x15\n\x10\x65NodeIsImportant\x10\x80\x01*\x08\x08\x64\x10\x80\x80\x80\x80\x02\":\n\x04Node\x12\x13\n\x04info\x18\x01 \x02(\x0b\x32\x05.Info\x12\x13\n\x04node\x18\x02 \x03(\x0b\x32\x05.Node*\x08\x08\x64\x10\x80\x80\x80\x80\x02\"Y\n\x08\x43hildAdd\x12\x16\n\x0eparent_node_id\x18\x01 \x02(\r\x12\x12\n\nchild_name\x18\x02 \x02(\t\x12\x17\n\x0f\x63hild_type_name\x18\x03 \x02(\t*\x08\x08\x64\x10\x80\x80\x80\x80\x02\"C\n\x0b\x43hildRemove\x12\x16\n\x0eparent_node_id\x18\x01 \x02(\r\x12\x12\n\nchild_name\x18\x02 \x02(\t*\x08\x08\x64\x10\x80\x80\x80\x80\x02\"\x92\x02\n\x0cVariantValue\x12\x0f\n\x07node_id\x18\x01 \x01(\r\x12\x0f\n\x07\x64_value\x18\x02 \x01(\x01\x12\x0f\n\x07\x66_value\x18\x03 \x01(\x02\x12\x12\n\nui64_value\x18\x04 \x01(\x04\x12\x11\n\ti64_value\x18\x05 \x01(\x12\x12\x10\n\x08ui_value\x18\x06 \x01(\r\x12\x0f\n\x07i_value\x18\x07 \x01(\x11\x12\x10\n\x08us_value\x18\x08 \x01(\r\x12\x0f\n\x07s_value\x18\t \x01(\x11\x12\x10\n\x08uc_value\x18\n \x01(\r\x12\x0f\n\x07\x63_value\x18\x0b \x01(\x11\x12\x0f\n\x07\x62_value\x18\x0c \x01(\x08\x12\x11\n\tstr_value\x18\r \x01(\t\x12\x11\n\ttimestamp\x18\x0e \x01(\x04*\x08\x08\x64\x10\x80\x80\x80\x80\x02\"C\n\x0cValueRequest\x12\x0f\n\x07node_id\x18\x01 \x02(\r\x12\n\n\x02\x66s\x18\x02 \x01(\x01\x12\x0c\n\x04stop\x18\x03 \x01(\x08*\x08\x08\x64\x10\x80\x80\x80\x80\x02*\xb9\x01\n\x0fRemoteErrorCode\x12\x14\n\x10\x65INVALID_REQUEST\x10\n\x12\x1f\n\x1b\x65UNSUPPORTED_CONTAINER_TYPE\x10\x14\x12\x1f\n\x1b\x65VALUE_THROTTLING_OCCURRING\x10\x1e\x12\x1d\n\x19\x65VALUE_THROTTLING_STOPPED\x10\x1f\x12\x15\n\x11\x65\x43HILD_ADD_FAILED\x10(\x12\x18\n\x14\x65\x43HILD_REMOVE_FAILED\x10\x32*\xfb\x01\n\x0b\x43\x44PNodeType\x12\x1a\n\rCDP_UNDEFINED\x10\xff\xff\xff\xff\xff\xff\xff\xff\xff\x01\x12\x0e\n\nCDP_SYSTEM\x10\x00\x12\x13\n\x0f\x43\x44P_APPLICATION\x10\x01\x12\x11\n\rCDP_COMPONENT\x10\x02\x12\x0e\n\nCDP_OBJECT\x10\x03\x12\x0f\n\x0b\x43\x44P_MESSAGE\x10\x04\x12\x13\n\x0f\x43\x44P_BASE_OBJECT\x10\x05\x12\x10\n\x0c\x43\x44P_PROPERTY\x10\x06\x12\x0f\n\x0b\x43\x44P_SETTING\x10\x07\x12\x0c\n\x08\x43\x44P_ENUM\x10\x08\x12\x10\n\x0c\x43\x44P_OPERATOR\x10\t\x12\x0c\n\x08\x43\x44P_NODE\x10\n\x12\x11\n\rCDP_USER_TYPE\x10\x64*\xbc\x01\n\x0c\x43\x44PValueType\x12\x0e\n\neUNDEFINED\x10\x00\x12\x0b\n\x07\x65\x44OUBLE\x10\x01\x12\x0b\n\x07\x65UINT64\x10\x02\x12\n\n\x06\x65INT64\x10\x03\x12\n\n\x06\x65\x46LOAT\x10\x04\x12\t\n\x05\x65UINT\x10\x05\x12\x08\n\x04\x65INT\x10\x06\x12\x0b\n\x07\x65USHORT\x10\x07\x12\n\n\x06\x65SHORT\x10\x08\x12\n\n\x06\x65UCHAR\x10\t\x12\t\n\x05\x65\x43HAR\x10\n\x12\t\n\x05\x65\x42OOL\x10\x0b\x12\x0b\n\x07\x65STRING\x10\x0c\x12\r\n\teUSERTYPE\x10\x64')
)

_REMOTEERRORCODE = _descriptor.EnumDescriptor(
  name='RemoteErrorCode',
  full_name='RemoteErrorCode',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='eINVALID_REQUEST', index=0, number=10,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='eUNSUPPORTED_CONTAINER_TYPE', index=1, number=20,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='eVALUE_THROTTLING_OCCURRING', index=2, number=30,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='eVALUE_THROTTLING_STOPPED', index=3, number=31,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='eCHILD_ADD_FAILED', index=4, number=40,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='eCHILD_REMOVE_FAILED', index=5, number=50,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=2658,
  serialized_end=2843,
)
_sym_db.RegisterEnumDescriptor(_REMOTEERRORCODE)

RemoteErrorCode = enum_type_wrapper.EnumTypeWrapper(_REMOTEERRORCODE)
_CDPNODETYPE = _descriptor.EnumDescriptor(
  name='CDPNodeType',
  full_name='CDPNodeType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='CDP_UNDEFINED', index=0, number=-1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CDP_SYSTEM', index=1, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CDP_APPLICATION', index=2, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CDP_COMPONENT', index=3, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CDP_OBJECT', index=4, number=3,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CDP_MESSAGE', index=5, number=4,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CDP_BASE_OBJECT', index=6, number=5,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CDP_PROPERTY', index=7, number=6,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CDP_SETTING', index=8, number=7,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CDP_ENUM', index=9, number=8,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CDP_OPERATOR', index=10, number=9,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CDP_NODE', index=11, number=10,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CDP_USER_TYPE', index=12, number=100,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=2846,
  serialized_end=3097,
)
_sym_db.RegisterEnumDescriptor(_CDPNODETYPE)

CDPNodeType = enum_type_wrapper.EnumTypeWrapper(_CDPNODETYPE)
_CDPVALUETYPE = _descriptor.EnumDescriptor(
  name='CDPValueType',
  full_name='CDPValueType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='eUNDEFINED', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='eDOUBLE', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='eUINT64', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='eINT64', index=3, number=3,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='eFLOAT', index=4, number=4,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='eUINT', index=5, number=5,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='eINT', index=6, number=6,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='eUSHORT', index=7, number=7,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='eSHORT', index=8, number=8,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='eUCHAR', index=9, number=9,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='eCHAR', index=10, number=10,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='eBOOL', index=11, number=11,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='eSTRING', index=12, number=12,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='eUSERTYPE', index=13, number=100,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=3100,
  serialized_end=3288,
)
_sym_db.RegisterEnumDescriptor(_CDPVALUETYPE)

CDPValueType = enum_type_wrapper.EnumTypeWrapper(_CDPVALUETYPE)
eINVALID_REQUEST = 10
eUNSUPPORTED_CONTAINER_TYPE = 20
eVALUE_THROTTLING_OCCURRING = 30
eVALUE_THROTTLING_STOPPED = 31
eCHILD_ADD_FAILED = 40
eCHILD_REMOVE_FAILED = 50
CDP_UNDEFINED = -1
CDP_SYSTEM = 0
CDP_APPLICATION = 1
CDP_COMPONENT = 2
CDP_OBJECT = 3
CDP_MESSAGE = 4
CDP_BASE_OBJECT = 5
CDP_PROPERTY = 6
CDP_SETTING = 7
CDP_ENUM = 8
CDP_OPERATOR = 9
CDP_NODE = 10
CDP_USER_TYPE = 100
eUNDEFINED = 0
eDOUBLE = 1
eUINT64 = 2
eINT64 = 3
eFLOAT = 4
eUINT = 5
eINT = 6
eUSHORT = 7
eSHORT = 8
eUCHAR = 9
eCHAR = 10
eBOOL = 11
eSTRING = 12
eUSERTYPE = 100


_AUTHRESPONSE_AUTHRESULTCODE = _descriptor.EnumDescriptor(
  name='AuthResultCode',
  full_name='AuthResponse.AuthResultCode',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='eUnknown', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='eGranted', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='eGrantedPasswordWillExpireSoon', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='eNewPasswordRequired', index=3, number=10,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='eInvalidChallengeResponse', index=4, number=11,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='eAdditionalResponseRequired', index=5, number=12,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=750,
  serialized_end=920,
)
_sym_db.RegisterEnumDescriptor(_AUTHRESPONSE_AUTHRESULTCODE)

_CONTAINER_TYPE = _descriptor.EnumDescriptor(
  name='Type',
  full_name='Container.Type',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='eRemoteError', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='eStructureRequest', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='eStructureResponse', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='eGetterRequest', index=3, number=3,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='eGetterResponse', index=4, number=4,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='eSetterRequest', index=5, number=5,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='eStructureChangeResponse', index=6, number=6,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='eCurrentTimeRequest', index=7, number=7,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='eCurrentTimeResponse', index=8, number=8,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='eChildAddRequest', index=9, number=9,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='eChildRemoveRequest', index=10, number=10,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=1327,
  serialized_end=1587,
)
_sym_db.RegisterEnumDescriptor(_CONTAINER_TYPE)

_INFO_FLAGS = _descriptor.EnumDescriptor(
  name='Flags',
  full_name='Info.Flags',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='eNone', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='eNodeIsLeaf', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='eValueIsPersistent', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='eValueIsReadOnly', index=3, number=4,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='eNodeIsRemovable', index=4, number=8,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='eNodeCanAddChildren', index=5, number=16,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='eNodeIsRenamable', index=6, number=32,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='eNodeIsInternal', index=7, number=64,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='eNodeIsImportant', index=8, number=128,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=1885,
  serialized_end=2079,
)
_sym_db.RegisterEnumDescriptor(_INFO_FLAGS)


_HELLO = _descriptor.Descriptor(
  name='Hello',
  full_name='Hello',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='system_name', full_name='Hello.system_name', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='compat_version', full_name='Hello.compat_version', index=1,
      number=2, type=13, cpp_type=3, label=2,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='incremental_version', full_name='Hello.incremental_version', index=2,
      number=3, type=13, cpp_type=3, label=2,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='public_key', full_name='Hello.public_key', index=3,
      number=4, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='challenge', full_name='Hello.challenge', index=4,
      number=5, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='application_name', full_name='Hello.application_name', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='cdp_version_major', full_name='Hello.cdp_version_major', index=6,
      number=7, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='cdp_version_minor', full_name='Hello.cdp_version_minor', index=7,
      number=8, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='cdp_version_patch', full_name='Hello.cdp_version_patch', index=8,
      number=9, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14,
  serialized_end=247,
)


_AUTHREQUEST_CHALLENGERESPONSE = _descriptor.Descriptor(
  name='ChallengeResponse',
  full_name='AuthRequest.ChallengeResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='AuthRequest.ChallengeResponse.type', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='response', full_name='AuthRequest.ChallengeResponse.response', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=342,
  serialized_end=393,
)

_AUTHREQUEST = _descriptor.Descriptor(
  name='AuthRequest',
  full_name='AuthRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='user_id', full_name='AuthRequest.user_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='challenge_response', full_name='AuthRequest.challenge_response', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_AUTHREQUEST_CHALLENGERESPONSE, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=250,
  serialized_end=393,
)


_ADDITIONALCHALLENGERESPONSEREQUIRED_PARAMETER = _descriptor.Descriptor(
  name='Parameter',
  full_name='AdditionalChallengeResponseRequired.Parameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='AdditionalChallengeResponseRequired.Parameter.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='AdditionalChallengeResponseRequired.Parameter.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=532,
  serialized_end=572,
)

_ADDITIONALCHALLENGERESPONSEREQUIRED = _descriptor.Descriptor(
  name='AdditionalChallengeResponseRequired',
  full_name='AdditionalChallengeResponseRequired',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='AdditionalChallengeResponseRequired.type', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='prompt', full_name='AdditionalChallengeResponseRequired.prompt', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='parameter', full_name='AdditionalChallengeResponseRequired.parameter', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_ADDITIONALCHALLENGERESPONSEREQUIRED_PARAMETER, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=396,
  serialized_end=572,
)


_AUTHRESPONSE = _descriptor.Descriptor(
  name='AuthResponse',
  full_name='AuthResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='result_code', full_name='AuthResponse.result_code', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='result_text', full_name='AuthResponse.result_text', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='additional_challenge_response_required', full_name='AuthResponse.additional_challenge_response_required', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _AUTHRESPONSE_AUTHRESULTCODE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=575,
  serialized_end=920,
)


_CONTAINER = _descriptor.Descriptor(
  name='Container',
  full_name='Container',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='message_type', full_name='Container.message_type', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='error', full_name='Container.error', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='structure_request', full_name='Container.structure_request', index=2,
      number=3, type=13, cpp_type=3, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='structure_response', full_name='Container.structure_response', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='getter_request', full_name='Container.getter_request', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='getter_response', full_name='Container.getter_response', index=5,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='setter_request', full_name='Container.setter_request', index=6,
      number=7, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='structure_change_response', full_name='Container.structure_change_response', index=7,
      number=8, type=13, cpp_type=3, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='current_time_response', full_name='Container.current_time_response', index=8,
      number=9, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='child_add_request', full_name='Container.child_add_request', index=9,
      number=10, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='child_remove_request', full_name='Container.child_remove_request', index=10,
      number=11, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _CONTAINER_TYPE,
  ],
  serialized_options=None,
  is_extendable=True,
  syntax='proto2',
  extension_ranges=[(100, 536870912), ],
  oneofs=[
  ],
  serialized_start=923,
  serialized_end=1597,
)


_ERROR = _descriptor.Descriptor(
  name='Error',
  full_name='Error',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='code', full_name='Error.code', index=0,
      number=1, type=13, cpp_type=3, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='text', full_name='Error.text', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='node_id', full_name='Error.node_id', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='parameter', full_name='Error.parameter', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=True,
  syntax='proto2',
  extension_ranges=[(100, 536870912), ],
  oneofs=[
  ],
  serialized_start=1599,
  serialized_end=1680,
)


_INFO = _descriptor.Descriptor(
  name='Info',
  full_name='Info',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='node_id', full_name='Info.node_id', index=0,
      number=1, type=13, cpp_type=3, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='name', full_name='Info.name', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='node_type', full_name='Info.node_type', index=2,
      number=3, type=14, cpp_type=8, label=2,
      has_default_value=False, default_value=-1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value_type', full_name='Info.value_type', index=3,
      number=4, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='type_name', full_name='Info.type_name', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='server_addr', full_name='Info.server_addr', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='server_port', full_name='Info.server_port', index=6,
      number=7, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='is_local', full_name='Info.is_local', index=7,
      number=8, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='flags', full_name='Info.flags', index=8,
      number=9, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _INFO_FLAGS,
  ],
  serialized_options=None,
  is_extendable=True,
  syntax='proto2',
  extension_ranges=[(100, 536870912), ],
  oneofs=[
  ],
  serialized_start=1683,
  serialized_end=2089,
)


_NODE = _descriptor.Descriptor(
  name='Node',
  full_name='Node',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='info', full_name='Node.info', index=0,
      number=1, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='node', full_name='Node.node', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=True,
  syntax='proto2',
  extension_ranges=[(100, 536870912), ],
  oneofs=[
  ],
  serialized_start=2091,
  serialized_end=2149,
)


_CHILDADD = _descriptor.Descriptor(
  name='ChildAdd',
  full_name='ChildAdd',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent_node_id', full_name='ChildAdd.parent_node_id', index=0,
      number=1, type=13, cpp_type=3, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='child_name', full_name='ChildAdd.child_name', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='child_type_name', full_name='ChildAdd.child_type_name', index=2,
      number=3, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=True,
  syntax='proto2',
  extension_ranges=[(100, 536870912), ],
  oneofs=[
  ],
  serialized_start=2151,
  serialized_end=2240,
)


_CHILDREMOVE = _descriptor.Descriptor(
  name='ChildRemove',
  full_name='ChildRemove',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent_node_id', full_name='ChildRemove.parent_node_id', index=0,
      number=1, type=13, cpp_type=3, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='child_name', full_name='ChildRemove.child_name', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=True,
  syntax='proto2',
  extension_ranges=[(100, 536870912), ],
  oneofs=[
  ],
  serialized_start=2242,
  serialized_end=2309,
)


_VARIANTVALUE = _descriptor.Descriptor(
  name='VariantValue',
  full_name='VariantValue',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='node_id', full_name='VariantValue.node_id', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='d_value', full_name='VariantValue.d_value', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='f_value', full_name='VariantValue.f_value', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ui64_value', full_name='VariantValue.ui64_value', index=3,
      number=4, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='i64_value', full_name='VariantValue.i64_value', index=4,
      number=5, type=18, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ui_value', full_name='VariantValue.ui_value', index=5,
      number=6, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='i_value', full_name='VariantValue.i_value', index=6,
      number=7, type=17, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='us_value', full_name='VariantValue.us_value', index=7,
      number=8, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='s_value', full_name='VariantValue.s_value', index=8,
      number=9, type=17, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='uc_value', full_name='VariantValue.uc_value', index=9,
      number=10, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='c_value', full_name='VariantValue.c_value', index=10,
      number=11, type=17, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='b_value', full_name='VariantValue.b_value', index=11,
      number=12, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='str_value', full_name='VariantValue.str_value', index=12,
      number=13, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='VariantValue.timestamp', index=13,
      number=14, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=True,
  syntax='proto2',
  extension_ranges=[(100, 536870912), ],
  oneofs=[
  ],
  serialized_start=2312,
  serialized_end=2586,
)


_VALUEREQUEST = _descriptor.Descriptor(
  name='ValueRequest',
  full_name='ValueRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='node_id', full_name='ValueRequest.node_id', index=0,
      number=1, type=13, cpp_type=3, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='fs', full_name='ValueRequest.fs', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='stop', full_name='ValueRequest.stop', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=True,
  syntax='proto2',
  extension_ranges=[(100, 536870912), ],
  oneofs=[
  ],
  serialized_start=2588,
  serialized_end=2655,
)

_AUTHREQUEST_CHALLENGERESPONSE.containing_type = _AUTHREQUEST
_AUTHREQUEST.fields_by_name['challenge_response'].message_type = _AUTHREQUEST_CHALLENGERESPONSE
_ADDITIONALCHALLENGERESPONSEREQUIRED_PARAMETER.containing_type = _ADDITIONALCHALLENGERESPONSEREQUIRED
_ADDITIONALCHALLENGERESPONSEREQUIRED.fields_by_name['parameter'].message_type = _ADDITIONALCHALLENGERESPONSEREQUIRED_PARAMETER
_AUTHRESPONSE.fields_by_name['result_code'].enum_type = _AUTHRESPONSE_AUTHRESULTCODE
_AUTHRESPONSE.fields_by_name['additional_challenge_response_required'].message_type = _ADDITIONALCHALLENGERESPONSEREQUIRED
_AUTHRESPONSE_AUTHRESULTCODE.containing_type = _AUTHRESPONSE
_CONTAINER.fields_by_name['message_type'].enum_type = _CONTAINER_TYPE
_CONTAINER.fields_by_name['error'].message_type = _ERROR
_CONTAINER.fields_by_name['structure_response'].message_type = _NODE
_CONTAINER.fields_by_name['getter_request'].message_type = _VALUEREQUEST
_CONTAINER.fields_by_name['getter_response'].message_type = _VARIANTVALUE
_CONTAINER.fields_by_name['setter_request'].message_type = _VARIANTVALUE
_CONTAINER.fields_by_name['child_add_request'].message_type = _CHILDADD
_CONTAINER.fields_by_name['child_remove_request'].message_type = _CHILDREMOVE
_CONTAINER_TYPE.containing_type = _CONTAINER
_INFO.fields_by_name['node_type'].enum_type = _CDPNODETYPE
_INFO.fields_by_name['value_type'].enum_type = _CDPVALUETYPE
_INFO_FLAGS.containing_type = _INFO
_NODE.fields_by_name['info'].message_type = _INFO
_NODE.fields_by_name['node'].message_type = _NODE
DESCRIPTOR.message_types_by_name['Hello'] = _HELLO
DESCRIPTOR.message_types_by_name['AuthRequest'] = _AUTHREQUEST
DESCRIPTOR.message_types_by_name['AdditionalChallengeResponseRequired'] = _ADDITIONALCHALLENGERESPONSEREQUIRED
DESCRIPTOR.message_types_by_name['AuthResponse'] = _AUTHRESPONSE
DESCRIPTOR.message_types_by_name['Container'] = _CONTAINER
DESCRIPTOR.message_types_by_name['Error'] = _ERROR
DESCRIPTOR.message_types_by_name['Info'] = _INFO
DESCRIPTOR.message_types_by_name['Node'] = _NODE
DESCRIPTOR.message_types_by_name['ChildAdd'] = _CHILDADD
DESCRIPTOR.message_types_by_name['ChildRemove'] = _CHILDREMOVE
DESCRIPTOR.message_types_by_name['VariantValue'] = _VARIANTVALUE
DESCRIPTOR.message_types_by_name['ValueRequest'] = _VALUEREQUEST
DESCRIPTOR.enum_types_by_name['RemoteErrorCode'] = _REMOTEERRORCODE
DESCRIPTOR.enum_types_by_name['CDPNodeType'] = _CDPNODETYPE
DESCRIPTOR.enum_types_by_name['CDPValueType'] = _CDPVALUETYPE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Hello = _reflection.GeneratedProtocolMessageType('Hello', (_message.Message,), dict(
  DESCRIPTOR = _HELLO,
  __module__ = 'cdp_pb2'
  # @@protoc_insertion_point(class_scope:Hello)
  ))
_sym_db.RegisterMessage(Hello)

AuthRequest = _reflection.GeneratedProtocolMessageType('AuthRequest', (_message.Message,), dict(

  ChallengeResponse = _reflection.GeneratedProtocolMessageType('ChallengeResponse', (_message.Message,), dict(
    DESCRIPTOR = _AUTHREQUEST_CHALLENGERESPONSE,
    __module__ = 'cdp_pb2'
    # @@protoc_insertion_point(class_scope:AuthRequest.ChallengeResponse)
    ))
  ,
  DESCRIPTOR = _AUTHREQUEST,
  __module__ = 'cdp_pb2'
  # @@protoc_insertion_point(class_scope:AuthRequest)
  ))
_sym_db.RegisterMessage(AuthRequest)
_sym_db.RegisterMessage(AuthRequest.ChallengeResponse)

AdditionalChallengeResponseRequired = _reflection.GeneratedProtocolMessageType('AdditionalChallengeResponseRequired', (_message.Message,), dict(

  Parameter = _reflection.GeneratedProtocolMessageType('Parameter', (_message.Message,), dict(
    DESCRIPTOR = _ADDITIONALCHALLENGERESPONSEREQUIRED_PARAMETER,
    __module__ = 'cdp_pb2'
    # @@protoc_insertion_point(class_scope:AdditionalChallengeResponseRequired.Parameter)
    ))
  ,
  DESCRIPTOR = _ADDITIONALCHALLENGERESPONSEREQUIRED,
  __module__ = 'cdp_pb2'
  # @@protoc_insertion_point(class_scope:AdditionalChallengeResponseRequired)
  ))
_sym_db.RegisterMessage(AdditionalChallengeResponseRequired)
_sym_db.RegisterMessage(AdditionalChallengeResponseRequired.Parameter)

AuthResponse = _reflection.GeneratedProtocolMessageType('AuthResponse', (_message.Message,), dict(
  DESCRIPTOR = _AUTHRESPONSE,
  __module__ = 'cdp_pb2'
  # @@protoc_insertion_point(class_scope:AuthResponse)
  ))
_sym_db.RegisterMessage(AuthResponse)

Container = _reflection.GeneratedProtocolMessageType('Container', (_message.Message,), dict(
  DESCRIPTOR = _CONTAINER,
  __module__ = 'cdp_pb2'
  # @@protoc_insertion_point(class_scope:Container)
  ))
_sym_db.RegisterMessage(Container)

Error = _reflection.GeneratedProtocolMessageType('Error', (_message.Message,), dict(
  DESCRIPTOR = _ERROR,
  __module__ = 'cdp_pb2'
  # @@protoc_insertion_point(class_scope:Error)
  ))
_sym_db.RegisterMessage(Error)

Info = _reflection.GeneratedProtocolMessageType('Info', (_message.Message,), dict(
  DESCRIPTOR = _INFO,
  __module__ = 'cdp_pb2'
  # @@protoc_insertion_point(class_scope:Info)
  ))
_sym_db.RegisterMessage(Info)

Node = _reflection.GeneratedProtocolMessageType('Node', (_message.Message,), dict(
  DESCRIPTOR = _NODE,
  __module__ = 'cdp_pb2'
  # @@protoc_insertion_point(class_scope:Node)
  ))
_sym_db.RegisterMessage(Node)

ChildAdd = _reflection.GeneratedProtocolMessageType('ChildAdd', (_message.Message,), dict(
  DESCRIPTOR = _CHILDADD,
  __module__ = 'cdp_pb2'
  # @@protoc_insertion_point(class_scope:ChildAdd)
  ))
_sym_db.RegisterMessage(ChildAdd)

ChildRemove = _reflection.GeneratedProtocolMessageType('ChildRemove', (_message.Message,), dict(
  DESCRIPTOR = _CHILDREMOVE,
  __module__ = 'cdp_pb2'
  # @@protoc_insertion_point(class_scope:ChildRemove)
  ))
_sym_db.RegisterMessage(ChildRemove)

VariantValue = _reflection.GeneratedProtocolMessageType('VariantValue', (_message.Message,), dict(
  DESCRIPTOR = _VARIANTVALUE,
  __module__ = 'cdp_pb2'
  # @@protoc_insertion_point(class_scope:VariantValue)
  ))
_sym_db.RegisterMessage(VariantValue)

ValueRequest = _reflection.GeneratedProtocolMessageType('ValueRequest', (_message.Message,), dict(
  DESCRIPTOR = _VALUEREQUEST,
  __module__ = 'cdp_pb2'
  # @@protoc_insertion_point(class_scope:ValueRequest)
  ))
_sym_db.RegisterMessage(ValueRequest)


# @@protoc_insertion_point(module_scope)



from promise import Promise
from time import sleep
from collections import namedtuple
from hashlib import sha256
import cdp_client.cdp_pb2 as proto
import websocket
import logging
import time

nanoseconds_in_second = 1000000000.0

def enum(**enums):
    return type('Enum', (), enums)

NodeType = enum(
    UNDEFINED=-1,
    SYSTEM=0,
    APPLICATION=1,
    COMPONENT=2,
    OBJECT=3,
    MESSAGE=4,
    BASE_OBJECT=5,
    PROPERTY=6,
    SETTING=7,
    ENUM=8,
    OPERATOR=9,
    NODE=10,
    USER_TYPE=100)


class ConnectionError(Exception):
    pass


class CommunicationError(Exception):
    pass


class InvalidRequestError(Exception):
    pass


class NotFoundError(Exception):
    pass


class UnknownError(Exception):
    pass


class Client:
    def __init__(self, host='127.0.0.1', port=7689, auto_reconnect=True, use_encryption=False, encryption_parameters=dict(),
                 user_id='', password=''):
        self._connection = Connection(host, port, auto_reconnect, use_encryption, encryption_parameters, user_id, password)

    def run_event_loop(self):
        self._connection.run_event_loop()

    def disconnect(self):
        self._connection.close()

    def root_node(self):
        return self._connection.node_tree().root_node()

    def find_node(self, path):
        def scan_node(node):
            tokens.pop(0)
            if not tokens:
                return Promise(lambda resolve, reject: resolve(node))
            return node.child(tokens[0]).then(scan_node)

        tokens = path.split('.')
        return self.root_node().then(scan_node)


class Node:
    def __init__(self, parent, connection, structure):
        self._connection = connection
        self._structure = structure
        self._children = []
        self._structure_subscriptions = []
        self._value_subscriptions = []
        self._value = proto.VariantValue()
        self._parent = parent
        for child in self._structure.node:
            self._children.append(Node(self, connection, child))

    def last_value(self):
        return self._value

    def set_value(self, value, timestamp=0):
        variant = self._value_to_variant(self._structure.info.value_type, value)
        variant.node_id = self._id()
        variant.timestamp = timestamp
        self._connection.send_value(variant)

    def name(self):
        return self._structure.info.name

    def path(self):
        return self.name() if self._parent is None else self._parent.path() + '.' + self.name()

    def parent(self):
        return self._parent

    def type(self):
        return self._translate_type(self._structure.info.node_type)

    def is_read_only(self):
        return self._structure.info.flags & proto.Info.eValueIsReadOnly != 0

    def is_leaf(self):
        return self._structure.info.flags & proto.Info.eNodeIsLeaf != 0

    def child(self, name):
        def update_node(node, structure):
            node._update_structure(structure)
            return Promise(lambda resolve, reject: resolve(node))

        for child in self._children:
            if child.name() == name:
                if child.is_leaf():
                    return Promise(lambda resolve, reject: resolve(child))
                return self._connection.send_structure_request(child._id(), child.path()).then(lambda structure: update_node(child, structure))
        return Promise(lambda resolve, reject: reject(NotFoundError("Could not find any children with name '" + name + "'")))

    def children(self):
        promises = []
        for child in self._children:
            promises.append(self.child(child.name()))
        return Promise.all(promises)

    def for_each_child(self, callback):
        for child in self._children:
            self.child(child.name()).done(callback)

    def subscribe_to_structure_changes(self, callback):
        self._structure_subscriptions.append(callback)

    def subscribe_to_value_changes(self, callback):
        self._value_subscriptions.append(callback)
        self._connection.send_value_request(self._id())

    def unsubscribe_from_structure_changes(self, callback):
        self._structure_subscriptions.remove(callback)

    def unsubscribe_from_value_changes(self, callback):
        self._value_subscriptions.remove(callback)
        if not self._value_subscriptions:
            self._connection.send_value_unrequest(self._id())

    def _id(self):
        return self._structure.info.node_id

    def _update(self):
        def update_structure(structure):
            self._update_structure(structure)
            return Promise(lambda resolve, reject: resolve(self))

        def fetch_structure():
            return self._connection.send_structure_request(self._id(), self.path())

        def fetch_value(node):
            if self._value_subscriptions:
                self._connection.send_value_request(self._id())
            return Promise(lambda resolve, reject: resolve(node))

        return fetch_structure().then(update_structure).then(fetch_value)

    def _update_structure(self, structure):
        self._structure = structure
        new_children = list(self._structure.node)
        lost_children = list(self._children)
        removed_children = []
        added_children = []

        def update_matching_children():
            for node in self._structure.node:
                for child in self._children:
                    if node.info.name == child.name():
                        child._structure = node

        def diff_children():
            for n in self._structure.node:
                for existing_child in self._children:
                    if n.info.node_id == existing_child._id():
                        new_children.remove(n)
                        lost_children.remove(existing_child)

        def report_children_diff():
            for child in lost_children:
                removed_children.append(child.name())
                self._children.remove(child)
            for child in new_children:
                node = Node(self, self._connection, child)
                self._children.append(node)
                added_children.append(node.name())

            if added_children or removed_children:
                for callback in self._structure_subscriptions:
                    callback(added_children, removed_children)

        update_matching_children()  # update children so that children structure response can lookup nodes by correct node id
        diff_children()
        report_children_diff()

    def _update_value(self, variant):
        self._value = self._value_from_variant(self._structure.info.value_type, variant)
        for callback in self._value_subscriptions:
            callback(self._value, variant.timestamp + self._connection.server_time_difference() * nanoseconds_in_second)

    @staticmethod
    def _translate_type(node_type):
        if node_type == proto.CDP_SYSTEM:
            return NodeType.SYSTEM
        elif node_type == proto.CDP_APPLICATION:
            return NodeType.APPLICATION
        elif node_type == proto.CDP_COMPONENT:
            return NodeType.COMPONENT
        elif node_type == proto.CDP_OBJECT:
            return NodeType.OBJECT
        elif node_type == proto.CDP_MESSAGE:
            return NodeType.MESSAGE
        elif node_type == proto.CDP_BASE_OBJECT:
            return NodeType.BASE_OBJECT
        elif node_type == proto.CDP_PROPERTY:
            return NodeType.PROPERTY
        elif node_type == proto.CDP_SETTING:
            return NodeType.SETTING
        elif node_type == proto.CDP_ENUM:
            return NodeType.ENUM
        elif node_type == proto.CDP_OPERATOR:
            return NodeType.OPERATOR
        elif node_type == proto.CDP_NODE:
            return NodeType.NODE
        elif node_type == proto.CDP_USER_TYPE:
            return NodeType.USER_TYPE
        else:
            return NodeType.UNDEFINED

    @staticmethod
    def _value_from_variant(type, variant):
        if type == proto.eDOUBLE:
            return variant.d_value
        elif type == proto.eUINT64:
            return variant.f_value
        elif type == proto.eINT64:
            return variant.ui64_value
        elif type == proto.eFLOAT:
            return variant.i64_value
        elif type == proto.eUINT:
            return variant.ui_value
        elif type == proto.eINT:
            return variant.i_value
        elif type == proto.eUSHORT:
            return variant.us_value
        elif type == proto.eSHORT:
            return variant.s_value
        elif type == proto.eUCHAR:
            return variant.uc_value
        elif type == proto.eCHAR:
            return variant.c_value
        elif type == proto.eBOOL:
            return variant.b_value
        elif type == proto.eSTRING:
            return variant.str_value
        else:
            return None

    @staticmethod
    def _value_to_variant(type, value):
        variant = proto.VariantValue()
        if type == proto.eDOUBLE:
            variant.d_value = value
        elif type == proto.eUINT64:
            variant.f_value = value
        elif type == proto.eINT64:
            variant.ui64_value = value
        elif type == proto.eFLOAT:
            variant.i64_value = value
        elif type == proto.eUINT:
            variant.ui_value = value
        elif type == proto.eINT:
            variant.i_value = value
        elif type == proto.eUSHORT:
            variant.us_value = value
        elif type == proto.eSHORT:
            variant.s_value = value
        elif type == proto.eUCHAR:
            variant.uc_value = value
        elif type == proto.eCHAR:
            variant.c_value = value
        elif type == proto.eBOOL:
            variant.b_value = value
        elif type == proto.eSTRING:
            variant.str_value = str(value)
        return variant


class Connection:
    def __init__(self, host, port, auto_reconnect, use_encryption=False, encryption_parameters=dict(),
                 user_id='', password=''):
        self._node_tree = NodeTree(self)
        self._structure_requests = Requests()
        self._time_request = Promise()
        self._time_diff = 0 #seconds
        self._last_time_diff_update = 0
        self._is_connected = False
        self._auto_reconnect = auto_reconnect
        self._use_encryption = use_encryption
        self._encryption_parameters = encryption_parameters
        self._challenge = ''
        self._user_id = user_id
        self._password = password
        if use_encryption:
            protocol = 'wss://'
        else:
            protocol = 'ws://'
        self._ws = self._connect(protocol + host + ":" + str(port))

    def node_tree(self):
        return self._node_tree

    def send_structure_request(self, node_id, node_path):
        p = Promise()
        self._structure_requests.add(node_path, p)
        if self._is_connected:
            self._update_time_difference()
            self._compose_and_send_structure_request(node_id)
        return p

    def send_value_request(self, node_id):
        self._update_time_difference()
        self._compose_and_send_value_request(node_id)

    def send_value_unrequest(self, node_id):
        self._update_time_difference()
        self._compose_and_send_value_request(node_id, True)

    def send_value(self, variant):
        self._update_time_difference()
        self._compose_and_send_value(variant)

    def run_event_loop(self):
        self._ws.run_forever(sslopt=self._encryption_parameters)
        while self._auto_reconnect:
            sleep(1)
            self._ws = self._connect(self._ws.url)
            self._ws.run_forever(sslopt=self._encryption_parameters)

    def close(self):
        self._auto_reconnect = False
        self._cleanup_queued_requests(ConnectionError('Connection was closed'))
        self._ws.close()

    def server_time_difference(self):
        return self._time_diff

    def _connect(self, url):
        return websocket.WebSocketApp(url,
                                      on_message=self._handle_hello_message,
                                      on_error=self._on_error,
                                      on_close=self._on_close,
                                      on_open=self._on_open)

    def _on_error(self, error):
        if not self._auto_reconnect:
            self._cleanup_queued_requests(ConnectionError(error))

    def _on_close(self):
        self._is_connected = False
        if not self._auto_reconnect:
            self._cleanup_queued_requests(ConnectionError("Connection was closed"))

    def _on_open(self):
        pass

    def _fetch_time_difference(self):
        def do_time_request():
            self._compose_and_send_time_request()
            return self._time_request

        def get_time_diff(time_samples):
            def get_best_sample():
                best_sample = None
                for sample in time_samples:
                    if best_sample is None or sample.ping < best_sample.ping:
                        best_sample = sample
                return best_sample
            return Promise(lambda resolve, reject: resolve(get_best_sample().diff))

        def calculate_time_diff(time_request_sent, response):
            time_response_received = time.time()
            client_time = time_response_received
            ping_time = time_response_received - time_request_sent
            server_time = response / nanoseconds_in_second + ping_time / 2.0
            time_diff = client_time - server_time  # time_diff in seconds
            Sample = namedtuple('Sample', 'ping, diff')
            return Promise(lambda resolve, reject: resolve(Sample(ping_time, time_diff)))

        def get_time_sample():
            time_request_sent = time.time()
            return do_time_request().then(lambda response: calculate_time_diff(time_request_sent, response))

        def get_time_samples():
            number_of_samples = 3
            time_samples = []
            promise = Promise()

            def get_more_samples_if_needed():
                def store_sample_and_get_more_if_needed(sample):
                    time_samples.append(sample)
                    return get_more_samples_if_needed()

                if len(time_samples) < number_of_samples:
                    return get_time_sample().then(store_sample_and_get_more_if_needed)
                else:
                    promise.do_resolve(time_samples)

            get_more_samples_if_needed()
            return promise

        return get_time_samples().then(get_time_diff)

    def _update_time_difference(self):
        maximum_time_diff_update_frequency = 10 #seconds

        def store_time_diff(time_diff):
            self._time_diff = time_diff
            self._last_time_diff_update = time.time()
            return Promise(lambda resolve, reject: resolve())

        if time.time() >= self._last_time_diff_update + maximum_time_diff_update_frequency:
            return self._fetch_time_difference().then(store_time_diff)
        else:
            return Promise(lambda resolve, reject: resolve())

    def _sync_time(self):
        self._is_connected = True
        self._ws.on_message = self._handle_container_message
        self._update_time_difference().then(self._node_tree.update()).then(self._send_queued_requests())

    def _handle_auth_response(self, message):
        data = proto.AuthResponse()
        data.ParseFromString(message)
        if data.result_code in (data.eGranted, data.eGrantedPasswordWillExpireSoon):
            self._sync_time()
        else:
            self._cleanup_queued_requests(CommunicationError(data.result_text))

    def _authenticate(self):
        self._ws.on_message = self._handle_auth_response
        self._compose_and_send_auth_request()

    def _handle_hello_message(self, message):
        if self._parse_hello_message(message):
            if self._challenge:
                self._authenticate()
            else:
                self._sync_time()
        else:
            self._cleanup_queued_requests(CommunicationError('Protocol mismatch'))

    def _handle_container_message(self, message):
        data = proto.Container()
        data.ParseFromString(message)
        if data.message_type == proto.Container.eStructureResponse:
            self._parse_structure_response(data.structure_response)
        elif data.message_type == proto.Container.eGetterResponse:
            self._parse_getter_response(data.getter_response)
        elif data.message_type == proto.Container.eStructureChangeResponse:
            self._parse_structure_change_response(data.structure_change_response)
        elif data.message_type == proto.Container.eCurrentTimeResponse:
            self._parse_current_time_response(data.current_time_response)
        elif data.message_type == proto.Container.eRemoteError:
            self._parse_error_response(data.error)
        else:
            logging.info('Unsupported message type received')

    def _parse_getter_response(self, response):
        for variant in response:
            node = self._node_tree.find_by_id(variant.node_id)
            node._update_value(variant)

    def _parse_structure_change_response(self, response):
        for node_id in response:
            node = self._node_tree.find_by_id(node_id)
            if node is not None:
                node._update()

    def _parse_current_time_response(self, response):
        self._time_request.do_resolve(response)

    def _parse_error_response(self, error):
        if error.code == proto.eINVALID_REQUEST:
            self._cleanup_queued_requests(InvalidRequestError(error.text))
        elif error.code == proto.eUNSUPPORTED_CONTAINER_TYPE:
            self._cleanup_queued_requests(CommunicationError(error.text))

    def _parse_hello_message(self, message):
        data = proto.Hello()
        data.ParseFromString(message)
        self._challenge = data.challenge
        if data.compat_version == 1:
            return True
        logging.info('Unsupported protocol version ' + str(data.compat_version) + '.' + str(data.incremental_version))
        return False

    def _parse_structure_response(self, response):
        for structure in response:
            node = self._node_tree.find_by_id(structure.info.node_id)
            node_path = node.path() if node is not None else None
            request = self._structure_requests.find(node_path)  # requests are stored with node path because node id can change between application reconnect
            if request is not None:
                self._structure_requests.remove(node_path)
                for p in request.promises:
                    p.do_resolve(structure)

    def _send_queued_requests(self):
        for request in self._structure_requests.get():
            node_path = request.node_path
            node_id = None if node_path is None else self._node_tree.find_by_path(node_path)._id()
            self._compose_and_send_structure_request(node_id)

    def _cleanup_queued_requests(self, error):
        self._time_request.reject(error)
        self._structure_requests.clear(error)

    def _compose_and_send_structure_request(self, node_id):
        data = proto.Container()
        data.message_type = proto.Container.eStructureRequest
        if node_id is not None:
            data.structure_request.append(node_id)
        self._ws.send(data.SerializeToString())

    def _compose_and_send_value_request(self, node_id, stop=False):
        data = proto.Container()
        data.message_type = proto.Container.eGetterRequest
        value = proto.ValueRequest()
        value.node_id = node_id
        value.fs = 5
        if stop:
            value.stop = stop
        data.getter_request.extend([value])
        self._ws.send(data.SerializeToString())

    def _compose_and_send_value(self, variant):
        data = proto.Container()
        data.message_type = proto.Container.eSetterRequest
        data.setter_request.extend([variant])
        self._ws.send(data.SerializeToString())

    def _compose_and_send_time_request(self):
        data = proto.Container()
        data.message_type = proto.Container.eCurrentTimeRequest
        self._ws.send(data.SerializeToString())

    def _compose_and_send_auth_request(self):
        request = proto.AuthRequest()
        request.user_id = self._user_id
        response = request.challenge_response.add()
        response.type = "PasswordHash"
        user_pass_hash = sha256(self._user_id.lower().encode() + b':' + self._password.encode()).digest()
        response.response = sha256(self._challenge + b':' + user_pass_hash).digest()
        self._ws.send(request.SerializeToString())


class NodeTree:
    def __init__(self, connection):
        self._connection = connection
        self._root_node = None  # starts with application node as node tree is created for each application connection

    def root_node(self):
        if self._root_node is None:
            return self._fetch_system() \
                .then(self._set_root_node) \
                .then(self._update_node)
        return Promise(lambda resolve, reject: resolve(self._root_node))

    def find_by_id(self, node_id):
        def find_node(node):
            if node._id() == node_id:
                return node
            for child in node._children:
                node = find_node(child)
                if node is not None:
                    return node
            return None

        if self._root_node is None:
            return None
        return find_node(self._root_node)

    def find_by_path(self, path):
        def find_node(node):
            tokens.pop(0)
            if not tokens:
                return node
            for child in node._children:
                if child.name() == tokens[0]:
                    return find_node(child)
            return None

        if self._root_node is None:
            return None
        tokens = path.split('.')
        return find_node(self._root_node)

    def update(self):
        if self._root_node is not None:
            return self._fetch_system() \
                .then(self._get_root_node) \
                .then(self._update_recursively)
        return Promise(lambda resolve, reject: resolve())

    def _update_recursively(self, node):
        def update_children(node):
            promises = []
            for child in node._children:
                if child._children or child.is_leaf():  # do not fetch more than needed
                    promises.append(self._update_recursively(child))
            return Promise.all(promises)

        return node._update().done(update_children)

    def _update_node(self, node):
        return node._update()

    def _set_root_node(self, system_structure):
        def find_local_app(apps):
            for app in apps:
                if app.info.is_local:
                    return app
            return None

        if not self._root_node:
            self._root_node = Node(None, self._connection, find_local_app(system_structure.node))
        return Promise(lambda resolve, reject: resolve(self._root_node))

    def _get_root_node(self, system_structure):
        return Promise(lambda resolve, reject: resolve(self._root_node))

    def _fetch_system(self):
        return self._connection.send_structure_request(None, None)


class Requests:
    def __init__(self):
        self._requests = []

    def get(self):
        return self._requests

    def add(self, node_path, promise):  # use node_path instead of node_id as this doesn't change after reconnect
        request = self.find(node_path)
        if request is None:
            Request = namedtuple('Request', 'node_path, promises')
            self._requests.append(Request(node_path, [promise]))
        else:
            if promise not in request.promises:
                request.promises.append(promise)

    def find(self, node_path):
        for r in self._requests:
            if r.node_path == node_path:
                return r
        return None

    def remove(self, node_path):
        request = self.find(node_path)
        self._requests.remove(request)

    def clear(self, error=UnknownError('Something has went wrong')):
        for request in self._requests:
            for p in request.promises:
                p.do_reject(error)
        del self._requests[:]



from cdp_client import cdp_pb2 as proto
from copy import copy
from hashlib import sha256

value1 = proto.VariantValue()
value1.node_id = 5
value1.d_value = 55
value1.timestamp = 777

value2 = proto.VariantValue()
value2.node_id = 5
value2.d_value = 66
value2.timestamp = 888

value1_node = proto.Node()
value1_node.info.node_id = 5
value1_node.info.name = "Value1"
value1_node.info.node_type = proto.CDP_PROPERTY
value1_node.info.value_type = proto.eDOUBLE
value1_node.info.flags = proto.Info.eNodeIsLeaf

comp1_node = proto.Node()
comp1_node.info.node_id = 9
comp1_node.info.name = "Comp1"
comp1_node.info.node_type = proto.CDP_COMPONENT
comp1_node.info.value_type = proto.eUNDEFINED
comp1_node.info.flags = proto.Info.eValueIsReadOnly

app1_node = proto.Node()
app1_node.info.node_id = 1
app1_node.info.name = "App1"
app1_node.info.node_type = proto.CDP_APPLICATION
app1_node.info.value_type = proto.eUNDEFINED
app1_node.info.is_local = True
app1_node.info.flags = proto.Info.eValueIsReadOnly | proto.Info.eNodeIsLeaf

app2_node = proto.Node()
app2_node.info.node_id = 2
app2_node.info.name = "App2"
app2_node.info.node_type = proto.CDP_APPLICATION
app2_node.info.value_type = proto.eUNDEFINED
app2_node.info.flags = proto.Info.eValueIsReadOnly

app3_node = proto.Node()
app3_node.info.node_id = 3
app3_node.info.name = "App3"
app3_node.info.node_type = proto.CDP_APPLICATION
app3_node.info.value_type = proto.eUNDEFINED
app3_node.info.flags = proto.Info.eValueIsReadOnly

system_node = proto.Node()
system_node.info.node_id = 0
system_node.info.name = "System"
system_node.info.node_type = proto.CDP_SYSTEM
system_node.info.value_type = proto.eUNDEFINED
system_node.info.flags = proto.Info.eValueIsReadOnly

hello_response = proto.Hello()
hello_response.system_name = "foo"
hello_response.compat_version = 1
hello_response.incremental_version = 0


def create_system_structure_response():
    response = proto.Container()
    response.message_type = proto.Container.eStructureResponse
    system = copy(system_node)
    system.node.extend([copy(app1_node), copy(app2_node)])
    response.structure_response.extend([system])
    return response


def create_app_structure_response():
    response = proto.Container()
    response.message_type = proto.Container.eStructureResponse
    app = copy(app2_node)
    app.node.extend([copy(value1_node)])
    response.structure_response.extend([app])
    return response


def create_structure_change_response(node_id):
    response = proto.Container()
    response.message_type = proto.Container.eStructureChangeResponse
    response.structure_change_response.extend([node_id])
    return response


def create_value_response():
    response = proto.Container()
    response.message_type = proto.Container.eGetterResponse
    response.getter_response.extend([value1])
    return response


def create_error_response():
    response = proto.Container()
    response.message_type = proto.Container.eRemoteError
    response.error.code = proto.eINVALID_REQUEST
    response.error.text = "foo"
    return response


def create_value_request(node_id):
    request = proto.Container()
    request.message_type = proto.Container.eGetterRequest
    value = proto.ValueRequest()
    value.node_id = node_id
    value.fs = 5
    request.getter_request.extend([value])
    return request


def create_value_unrequest(node_id):
    request = proto.Container()
    request.message_type = proto.Container.eGetterRequest
    value = proto.ValueRequest()
    value.node_id = node_id
    value.fs = 5
    value.stop = True
    request.getter_request.extend([value])
    return request


def create_time_request():
    request = proto.Container()
    request.message_type = proto.Container.eCurrentTimeRequest
    return request


def create_time_response(time):
    response = proto.Container()
    response.message_type = proto.Container.eCurrentTimeResponse
    response.current_time_response = time
    return response


def create_valid_hello_response():
    response = proto.Hello()
    response.system_name = "foo"
    response.compat_version = 1
    response.incremental_version = 0
    return response


def create_valid_hello_response_with_auth_required(challenge):
    response = proto.Hello()
    response.system_name = "foo"
    response.compat_version = 1
    response.incremental_version = 0
    response.challenge = challenge
    return response


def create_invalid_hello_response():
    response = proto.Hello()
    response.system_name = "foo"
    response.compat_version = 2
    response.incremental_version = 0
    return response


def create_structure_request(node_id = None):
    request = proto.Container()
    request.message_type = proto.Container.eStructureRequest
    if node_id is not None:
        request.structure_request.append(node_id)
    return request


def create_setter_request(variant_value):
    request = proto.Container()
    request.message_type = proto.Container.eSetterRequest
    request.setter_request.extend([variant_value])
    return request


def create_structure_change_request(node_id):
    request = proto.Container()
    request.message_type = proto.Container.eStructureRequest
    request.structure_request.append(node_id)
    return request


def create_password_auth_request(challenge, user_id, password):
    request = proto.AuthRequest()
    request.user_id = user_id
    response = request.challenge_response.add()
    response.type = "PasswordHash"
    user_pass_hash = sha256(user_id.encode().lower() + b':' + password.encode()).digest()
    response.response = sha256(challenge + b':' + user_pass_hash).digest()
    return request


def create_auth_response_granted():
    response = proto.AuthResponse()
    response.result_code = response.eGranted
    return response


def create_auth_response_denied():
    response = proto.AuthResponse()
    response.result_code = response.eInvalidChallengeResponse
    return response



from cdp_client import cdp
from cdp_client.tests import fake_data
from collections import namedtuple
import unittest
import mock


class ConnectionTester(unittest.TestCase):
    def __init__(self, method_name):
        unittest.TestCase.__init__(self, method_name)
        self._connection = None

    def setUp(self):
        self._connection = cdp.Connection("foo", "bar", False)

    def tearDown(self):
        del self._connection

    @mock.patch.object(cdp.Requests, 'add')
    @mock.patch.object(cdp.websocket.WebSocketApp, 'send')
    def test_sending_structure_request_when_not_connected(self, mock_send, mock_add):
        self._connection.send_structure_request(1, 'foo')
        self.assertTrue(mock_add.called)
        mock_send.assert_not_called()

    @mock.patch.object(cdp.Requests, 'add')
    @mock.patch.object(cdp.websocket.WebSocketApp, 'send')
    def test_sending_structure_request_when_connected(self, mock_send, mock_add):
        self._connection._is_connected = True
        node_id = 1
        self._connection.send_structure_request(node_id, 'foo')
        self.assertTrue(mock_add.called)
        mock_send.assert_any_call(fake_data.create_structure_request(node_id).SerializeToString())

    @mock.patch.object(cdp.Requests, 'add')
    @mock.patch.object(cdp.websocket.WebSocketApp, 'send')
    def test_sending_root_node_structure_request(self, mock_send, mock_add):
        self._connection._is_connected = True
        self._connection.send_structure_request(None, 'foo')
        self.assertTrue(mock_add.called)
        mock_send.assert_any_call(fake_data.create_structure_request().SerializeToString())

    @mock.patch.object(cdp.websocket.WebSocketApp, 'send')
    def test_sending_value_request(self, mock_send):
        node_id = 1
        self._connection.send_value_request(node_id)
        mock_send.assert_any_call(fake_data.create_value_request(node_id).SerializeToString())

    @mock.patch.object(cdp.websocket.WebSocketApp, 'send')
    def test_sending_value_unrequest(self, mock_send):
        node_id = 1
        self._connection.send_value_unrequest(node_id)
        mock_send.assert_any_call(fake_data.create_value_unrequest(node_id).SerializeToString())

    @mock.patch.object(cdp.websocket.WebSocketApp, 'send')
    def test_sending_value(self, mock_send):
        self._connection.send_value(fake_data.value1)
        mock_send.assert_any_call(fake_data.create_setter_request(fake_data.value1).SerializeToString())

    @mock.patch.object(cdp.websocket.WebSocketApp, 'run_forever')
    def test_run_event_loop(self, mock_run_forever):
        self._connection.run_event_loop()
        mock_run_forever.assert_called_once_with(sslopt=dict())

    @mock.patch.object(cdp.websocket.WebSocketApp, 'close')
    def test_closing(self, mock_close):
        self._connection.close()
        mock_close.assert_called_once_with()

    @mock.patch.object(cdp.websocket.WebSocketApp, 'send')
    def test_connected_state_is_set_when_receiving_hello_message_with_correct_version(self, mock_send):
        self.assertEquals(self._connection._is_connected, False)
        self._connection._handle_hello_message(fake_data.create_valid_hello_response().SerializeToString())
        self.assertEquals(self._connection._is_connected, True)

    def test_connected_state_is_unset_when_receiving_hello_message_with_incorrect_version(self):
        self.assertEquals(self._connection._is_connected, False)
        self._connection._handle_hello_message(fake_data.create_invalid_hello_response().SerializeToString())
        self.assertEquals(self._connection._is_connected, False)

    @mock.patch.object(cdp.NodeTree, 'find_by_id')
    @mock.patch.object(cdp.Node, '_update_value')
    def test_node_updated_when_node_value_received(self, mock_update_value, mock_find_by_id):
        self._connection._is_connected = True
        mock_find_by_id.return_value = cdp.Node(None, self._connection, fake_data.app1_node)
        response = fake_data.create_value_response()
        self._connection._handle_container_message(response.SerializeToString())
        mock_update_value.assert_called_once_with(response.getter_response[0])

    @mock.patch.object(cdp.NodeTree, 'find_by_id')
    @mock.patch.object(cdp.websocket.WebSocketApp, 'send')
    def test_node_structure_requested_when_node_structure_change_received(self, mock_send, mock_find_by_id):
        self._connection._is_connected = True
        mock_find_by_id.return_value = cdp.Node(None, self._connection, fake_data.app1_node)
        response = fake_data.create_structure_change_response(fake_data.app1_node.info.node_id)
        request = fake_data.create_structure_change_request(response.structure_change_response[0])
        self._connection._handle_container_message(response.SerializeToString())
        mock_send.assert_any_call(request.SerializeToString())

    @mock.patch.object(cdp.Requests, 'clear')
    def test_requests_cleared_when_error_received(self, mock_clear):
        self._connection._handle_container_message(fake_data.create_error_response().SerializeToString())
        self.assertTrue(mock_clear.called)

    @mock.patch.object(cdp.websocket.WebSocketApp, 'send')
    @mock.patch.object(cdp.time, 'time')
    def _test_time_difference_udpate(self, mock_time, mock_send):
        def side_effect(*args, **kwargs):
            nsec_in_sec = 1000000000
            sample = samples.pop(0)
            mock_time.side_effect = [initial_time + sample.ping, initial_time]
            self._connection._handle_container_message(fake_data.create_time_response(sample.diff * nsec_in_sec).SerializeToString())

        initial_time = 10
        mock_time.return_value = initial_time
        mock_send.side_effect = side_effect
        samples = []

        Sample = namedtuple('Sample', 'ping, diff')
        samples.append(Sample(20, 100))
        samples.append(Sample(10, 200)) #this should be selected as it has the best ping
        samples.append(Sample(30, 300))

        self._connection._update_time_difference()
        request = fake_data.create_time_request().SerializeToString()
        mock_send.assert_has_calls([mock.call(request), mock.call(request), mock.call(request)])
        self.assertEqual(self._connection.server_time_difference(), -185.0)

    @mock.patch.object(cdp.websocket.WebSocketApp, 'send')
    @mock.patch.object(cdp.time, 'time')
    def test_time_difference_udpate_frequency(self, mock_time, mock_send):
        def side_effect(*args, **kwargs):
            self._connection._handle_container_message(fake_data.create_time_response(1).SerializeToString())

        mock_time.return_value = 10
        mock_send.side_effect = side_effect
        self._connection._update_time_difference()
        self.assertTrue(mock_send.called)
        mock_send.reset_mock()
        self._connection._update_time_difference()
        self.assertFalse(mock_send.called)

    @mock.patch.object(cdp.websocket.WebSocketApp, 'send')
    def test_password_auth_request_is_sent_when_required_by_server(self, mock_send):
        self.assertEquals(self._connection._is_connected, False)
        self._connection._user_id = "Testuser"
        self._connection._password = "testpass"
        self._connection._handle_hello_message(fake_data.create_valid_hello_response_with_auth_required(b'challenge')
                                               .SerializeToString())
        self.assertTrue(mock_send.called)
        mock_send.assert_any_call(fake_data.create_password_auth_request(self._connection._challenge,
                                                                         self._connection._user_id,
                                                                         self._connection._password).SerializeToString())
        self.assertEquals(self._connection._is_connected, False)

    @mock.patch.object(cdp.websocket.WebSocketApp, 'send')
    def test_connected_state_is_set_when_receiving_auth_response_granted(self, mock_send):
        self.assertEquals(self._connection._is_connected, False)
        self._connection._handle_auth_response(fake_data.create_auth_response_granted().SerializeToString())
        self.assertEquals(self._connection._is_connected, True)

    @mock.patch.object(cdp.websocket.WebSocketApp, 'send')
    def test_connected_state_is_set_when_receiving_auth_response_invalid(self, mock_send):
        self.assertEquals(self._connection._is_connected, False)
        self._connection._handle_auth_response(fake_data.create_auth_response_denied().SerializeToString())
        self.assertEquals(self._connection._is_connected, False)



from cdp_client import cdp
from promise import Promise
import unittest


class RequestsTester(unittest.TestCase):
    def __init__(self, method_name):
        unittest.TestCase.__init__(self, method_name)
        self._requests = None

    def setUp(self):
        self._requests = cdp.Requests()

    def tearDown(self):
        del self._requests

    def test_adding_one_request(self):
        self._requests.add(1, 2)
        self.assertEqual(len(self._requests.get()), 1)
        self.assertEqual(self._requests.get()[0].node_path, 1)
        self.assertEqual(self._requests.get()[0].promises, [2])

    def test_adding_same_requests(self):
        self._requests.add(1, 2)
        self._requests.add(1, 2)
        self.assertEqual(len(self._requests.get()), 1)
        self.assertEqual(self._requests.get()[0].node_path, 1)
        self.assertEqual(self._requests.get()[0].promises, [2])

    def test_adding_requests_with_same_node_id(self):
        self._requests.add(1, 2)
        self._requests.add(1, 4)
        self.assertEqual(len(self._requests.get()), 1)
        self.assertEqual(len(self._requests.get()[0].promises), 2)
        self.assertEqual(self._requests.get()[0].node_path, 1)
        self.assertEqual(self._requests.get()[0].promises, [2, 4])

    def test_adding_different_requests(self):
        self._requests.add(1, 2)
        self._requests.add(2, 4)
        self.assertEqual(len(self._requests.get()), 2)
        self.assertEqual(len(self._requests.get()[0].promises), 1)
        self.assertEqual(self._requests.get()[0].node_path, 1)
        self.assertEqual(self._requests.get()[0].promises, [2])
        self.assertEqual(self._requests.get()[1].node_path, 2)
        self.assertEqual(self._requests.get()[1].promises, [4])

    def test_finding_of_node(self):
        self._requests.add(1, 2)
        self._requests.add(2, 4)
        self._requests.add(3, 4)
        request = self._requests.find(2)
        self.assertEqual(request.node_path, 2)
        self.assertEqual(request.promises, [4])

    def test_removal_of_node(self):
        self._requests.add(1, 2)
        self._requests.add(2, 4)
        self._requests.add(2, 6)
        self._requests.add(3, 8)
        self._requests.remove(2)
        self.assertEqual(len(self._requests.get()), 2)
        self.assertEqual(self._requests.find(2), None)

    def test_clearing_all_requests(self):
        actual_errors = []
        expected_error = cdp.ConnectionError('foo')
        p1 = Promise()
        p2 = Promise()
        p1.catch(lambda error: actual_errors.append(error))
        p2.catch(lambda error: actual_errors.append(error))
        self._requests.add(1, p1)
        self._requests.add(2, p2)
        self._requests.clear(expected_error)
        self.assertEqual(len(self._requests.get()), 0)
        self.assertEqual(len(actual_errors), 2)
        self.assertEqual(actual_errors[0], expected_error)
        self.assertEqual(actual_errors[1], expected_error)



from cdp_client import cdp
from cdp_client.tests import fake_data
from promise import Promise
from copy import copy
import unittest
import mock


class NodeTreeTester(unittest.TestCase):
    def __init__(self, method_name):
        unittest.TestCase.__init__(self, method_name)
        self._node_tree = None

    def setUp(self):
        self._node_tree = cdp.NodeTree(cdp.Connection("foo", "bar", False))

    def tearDown(self):
        del self._node_tree

    @mock.patch.object(cdp.Connection, 'send_structure_request')
    def test_fetching_root_node_when_root_node_not_existing(self, mock_send_structure_request):
        nodes = []
        system_node = copy(fake_data.system_node)
        app_node = copy(fake_data.app1_node)
        app_node.node.extend([fake_data.value1_node])
        system_node.node.extend([app_node])
        mock_send_structure_request.side_effect = [Promise(lambda resolve, reject: resolve(system_node)),
                                                   Promise(lambda resolve, reject: resolve(app_node))]
        self._node_tree.root_node().then(lambda n: nodes.append(n))
        self.assertEquals(len(nodes), 1)
        self.assertEquals(nodes[0]._id(), app_node.info.node_id)

    @mock.patch.object(cdp.Connection, 'send_structure_request')
    def test_fetching_root_node_when_root_node_existing(self, mock_send_structure_request):
        nodes = []
        root_node = cdp.Node(None, self._node_tree._connection, fake_data.app1_node)
        self._node_tree._root_node = root_node
        self._node_tree.root_node().then(lambda n: nodes.append(n))
        mock_send_structure_request.assert_not_called()
        self.assertEquals(len(nodes), 1)
        self.assertEquals(nodes[0]._id(), root_node._id())

    def test_find_node_by_id(self):
        system_node = copy(fake_data.system_node)
        app_node = copy(fake_data.app1_node)
        app_node.node.extend([fake_data.value1_node])
        system_node.node.extend([app_node])
        root_node = cdp.Node(None, self._node_tree._connection, system_node)
        self._node_tree._root_node = root_node
        node = self._node_tree.find_by_id(fake_data.value1_node.info.node_id)
        self.assertEquals(node._id(), fake_data.value1_node.info.node_id)

    def test_find_node_by_path(self):
        system_node = copy(fake_data.system_node)
        app_node = copy(fake_data.app1_node)
        app_node.node.extend([fake_data.value1_node])
        system_node.node.extend([app_node])
        root_node = cdp.Node(None, self._node_tree._connection, system_node)
        self._node_tree._root_node = root_node
        path = '.'.join([system_node.info.name, app_node.info.name, fake_data.value1_node.info.name])
        node = self._node_tree.find_by_path(path)
        self.assertEquals(node._id(), fake_data.value1_node.info.node_id)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
from setuptools import setup, find_packages

def readme():
    with open('README.rst') as f:
        return f.read()

setup(
    name='cdp-client',
    version='1.2.1',
    packages=find_packages(),
    install_requires=[
        'promise==2.2.1',
        'websocket-client==0.56.0',
        'protobuf==3.7.1',
        'mock==3.0.5'],
    keywords=["cdp cdpstudio studio client cdp-client cdp_client"],
    url='https://github.com/CDPTechnologies/PythonCDPClient',
    license='MIT',
    author='CDP Technologies AS',
    author_email='info@cdptech.com',
    description='Provides an API that allows to interact with CDP applications',
	long_description=readme(),
    test_suite='nose.collector',
    tests_require=['nose'],
    include_package_data=True,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Topic :: Software Development',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6']
)
